#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>
#include <X11/Xlib.h>
#include <alsa/asoundlib.h>
#include <alsa/mixer.h>
#include <getopt.h>

#include <sys/stat.h> //qamix or alsamixer 

#include "tray.h"

// variable

#define ICON_PATH ICON_DIR "/tray_mixer_plus/"
#define RUN_AM "xterm -T AlsaMixer -fn 10 -e alsamixer"
#define RUN_GNOME "/usr/bin/gnome-alsamixer"
#define RUN_QAMIX "/usr/bin/qamix"
#define RUN_GMIXER "/usr/bin/gmixer"
#define RUN_AMGUI "/usr/bin/alsamixergui"
#define CHANGE_SOUNDCARD "/usr/bin/asoundconf-gtk"
#define TEST_SOUNDCARD "/usr/bin/speaker-test"
#define TEST_SOUNDCARD_RUN "xterm -T TestSoundCard -fn 10 -bg black -cr red -fg yellow  -geometry 88x37 -e 'echo  ; echo 2.0 OUTPUT TEST; speaker-test -c2 -twav -l1; echo ; echo ; echo 5.1 OUTPUT TEST; speaker-test -c6 -twav -l1'"
#define DEF_ELEM_1 "Master"
#define DEF_ELEM_2 "Headphone"
#define DEF_ELEM_3 "PCM"

struct channel {
	GtkWidget *vbox;
	GtkWidget *vscale;
	GtkWidget *mute;
	GtkWidget *label;
	int muteval;
};

GtkStatusIcon *icon;
GtkWidget *window;
GtkWidget *hbox;
GtkWidget *menu, *item, *item2, *item3, *item4, *sep, *item_i, *item2_i, *item3_i, *item4_i;
struct channel ch[1];
snd_mixer_t *mixer;
snd_mixer_elem_t *elem;
snd_mixer_selem_id_t *sid;
int show=0;
int mytimertime=-1;
int wm=0;
char *name = "Master";
int userelement=0;
int dynamic=0;
int find_sc=0;

static gboolean on_mixer_event(GIOChannel* channel, GIOCondition cond, void *ud);
static void update_gui(struct channel *c);



static int mixer_init()
{
	int n=0, i=0, j=0, f=0;
	struct pollfd *fds;

	snd_mixer_selem_id_alloca(&sid);
	if (snd_mixer_open(&mixer, 0) < 0) {
		fprintf(stderr, "can't open mixer\n");
		return 1;
	}
	snd_mixer_attach(mixer, "default");
	snd_mixer_selem_register(mixer, NULL, NULL);
	snd_mixer_load(mixer);
	
	if (userelement==0) { j=3; } else { j=1; };
	
	for (i = 0; i < j; i++) {
	    if (userelement==0) {
		if (i==0) name=DEF_ELEM_1;
		if (i==1) name=DEF_ELEM_2;
		if (i==2) name=DEF_ELEM_3;
	    };
	    snd_mixer_selem_id_set_name(sid, name);
	    elem = snd_mixer_find_selem(mixer, sid);
	    if (elem) {
			g_print("find PCM element: %s\n", name);
			f++;
			break;
	    }
	} // for
	if (f==0) {
		fprintf(stderr, "can't find PCM element\n");
		return 1;
	}

	snd_mixer_selem_set_playback_volume_range(elem, 0, 100);

	n = snd_mixer_poll_descriptors_count(mixer);
	fds = calloc(n, sizeof(struct pollfd));
	snd_mixer_poll_descriptors(mixer, fds, n);

	for (i = 0; i < n; i++) {
		GIOChannel* channel = g_io_channel_unix_new( fds[i].fd );
		g_io_add_watch(channel, G_IO_IN|G_IO_HUP, on_mixer_event, NULL);
		g_io_channel_unref(channel);
	}

	return 0;
}

static int reread_mixer_init()
{
    int mi_ret=0;
    mi_ret=mixer_init();
    char name_gtk[10];

    if (mi_ret==0)
    {
	find_sc=1;
	gtk_status_icon_set_visible(GTK_STATUS_ICON(icon), TRUE);
	snprintf(name_gtk, 10, "%s", name);
	gtk_label_set_text( (GtkLabel *) ch[0].label, name_gtk);
    }else{
	return 1;
    }
    return 0;
}


static int mixer_get(struct channel *c)
{
	long l, r;

	snd_mixer_selem_get_playback_volume(elem,
				SND_MIXER_SCHN_FRONT_LEFT, &l);
	snd_mixer_selem_get_playback_volume(elem,
				SND_MIXER_SCHN_FRONT_RIGHT, &r);

	return (l + r) / 2;
}

static void mixer_set(struct channel *c, int vol)
{
	snd_mixer_selem_set_playback_volume(elem,
				SND_MIXER_SCHN_FRONT_LEFT, vol);
	snd_mixer_selem_set_playback_volume(elem,
				SND_MIXER_SCHN_FRONT_RIGHT, vol);
}

static void mixer_getmute(struct channel *c)
{
	if (c->mute) {
		snd_mixer_selem_get_playback_switch(elem,
				SND_MIXER_SCHN_FRONT_LEFT, &c->muteval);
	} else {
		c->muteval = 1;
	}
}

gboolean mixer_evt_idle;

static gboolean reset_mixer_evt_idle()
{
    mixer_evt_idle = 0;
    return FALSE;
}

static gboolean on_mixer_event(GIOChannel* channel, GIOCondition cond, void *ud)
{
	if (mixer_evt_idle == 0) {
		mixer_evt_idle = g_idle_add_full(G_PRIORITY_DEFAULT,
						 (GSourceFunc) reset_mixer_evt_idle,
						 NULL, NULL);
		snd_mixer_handle_events (mixer);
	}

	if (cond & G_IO_IN) {
		/* update mixer status */
		update_gui (&ch[0]);
	}

	if (cond & G_IO_HUP) {
		/* FIXME: This means there're some problems with alsa. */
		return FALSE;
	}

	return TRUE;
}

static void update_icon(struct channel *c)
{
        gtk_status_icon_set_from_file(icon, c->muteval ?
			ICON_PATH "speaker.png" : ICON_PATH "mute.png");
			
	char txt[15];
	snprintf(txt, 15, "%s %i %%", name,  (int) (  gtk_range_get_value((GtkRange *) c->vscale)  ) );
	gtk_status_icon_set_tooltip(icon, txt);
}

static void update_gui(struct channel *c)
{
	mixer_getmute(c);
	update_icon(c);

	if (c->mute) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(c->mute),
					!c->muteval);
	} else {
		c->muteval = 1;
	}

	gtk_range_set_value(GTK_RANGE(c->vscale), mixer_get(c));
}

static void vol_change(GtkRange *range, struct channel *c)
{
	mixer_set(c, gtk_range_get_value(range));
}

static void mute(GtkWidget *widget, struct channel *c)
{
	int val, i;

	val = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
	for (i = 0; i <= SND_MIXER_SCHN_LAST; i++)
		snd_mixer_selem_set_playback_switch(elem, i, !val);

	mixer_getmute(c);
	update_icon(c);
}

static int scale_scroll(GtkScale* scale, GdkEventScroll *e, struct channel *c)
{
	int val = gtk_range_get_value((GtkRange*)scale);

	val += e->direction == GDK_SCROLL_UP ? 2 : -2;
	gtk_range_set_value((GtkRange*)scale, CLAMP(val, 0, 100));

	return 0;
}


gboolean mytimer(gpointer data)
{
    mytimertime=mytimertime-1;
//    if (mytimertime>0) {g_timeout_add(1000, mytimer, NULL);}
//g_print("timer  %i\n", mytimertime);
    if (mytimertime==0){
	//hide window
	show = 0;
	mytimertime=-1;
        gtk_widget_hide(window);
	gtk_widget_hide_all(window);
    }
    if (mytimertime<0){
	//stop 
	return FALSE;
    }

    return TRUE;
}


gboolean mytimer2(gpointer data)
{
    reread_mixer_init();
    if (find_sc==1){
    	return FALSE;
    }
    return TRUE;
}



static void click()
{
	int forceshow=0;
	int ok=1;
	
	if (dynamic==1 && show==0){
	    if (reread_mixer_init()==1) {
		ok=0;
		gtk_status_icon_set_visible(GTK_STATUS_ICON(icon), FALSE);
		//run timer
		g_timeout_add(1000, mytimer2, NULL);
	    }
	}
    if (ok==1){

	if (show==1) {
	    gtk_widget_hide(window);
	    gtk_widget_hide_all(window);
		mytimertime=-1;
		show=0;
	} else {
		forceshow=1;
	}
	
	if (forceshow==1){
		if (wm==0){
		gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_MOUSE);
		}else{
		gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
		}
		gtk_window_stick(GTK_WINDOW(window));
		gtk_widget_show_all(window);
		
		//run timer
		mytimertime=10;
		g_timeout_add(1000, mytimer, NULL);
		show=1;	
	}
	
    }//if (ok==1){
}

static void popup()
{
		gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, 3,
						gtk_get_current_event_time());
}


static void cursor_enter()
{
    //
    mytimertime=100;
//g_print("cursor_enter %i\n", mytimertime);
}

static void cursor_leave()
{
    int wh, ww, px, py;
    
    gtk_window_get_size( (GtkWindow *) window, &ww, &wh);
    gtk_widget_get_pointer( (GtkWidget *) window, &px, &py);
    
    if (px<1 || px>ww-1  ||  py<1 || py>wh+1 )  { mytimertime=3; };

//g_print("px=%i py=%i  ww=%i wh=%i\n", px, py,  ww, wh );

//g_print("cursor_leave %i\n", mytimertime);

}

static void quit()
{
	gtk_main_quit();
}

static void extra()
{
    //run extra volume control
    struct stat st;
    int zzz;
    if ( stat(RUN_GNOME, &st)==0 ) {
	zzz=system(RUN_GNOME);
    }else{
	if ( stat(RUN_QAMIX, &st)==0 ) {
	    zzz=system(RUN_QAMIX);
	}else{
	    if ( stat(RUN_GMIXER, &st)==0 ) {
		zzz=system(RUN_GMIXER);
	    }else{
		if ( stat(RUN_AMGUI, &st)==0 ) {
		    zzz=system(RUN_AMGUI);
		}else{
		    //DEFAULT
		    zzz=system(RUN_AM);
		}
	    }
	}
    }

}


static void testsound()
{
    //run testsound
    struct stat st;
    int zzz;
    if ( stat(TEST_SOUNDCARD, &st)==0 ) {
	zzz=system(TEST_SOUNDCARD_RUN);
    }
}


static void defsound()
{
    //chande default sound card
    struct stat st;
    int zzz;
    if ( stat(CHANGE_SOUNDCARD, &st)==0 ) {
	zzz=system(CHANGE_SOUNDCARD);
    }
    //if ( dynamic==1 ){
	zzz=reread_mixer_init();
    //}
}


int main(int argc, char **argv)
{
	int o=0, noquit=0, unmute=0, i=0, unvoldig=0, sleep=0, mi_ret=0;
	char *unvol = "";
	char name_gtk[10];

	bindtextdomain("tray_mixer_plus", LOCALE_DIR);
	textdomain("tray_mixer_plus");

	wm=0; //center window position = no
	
	while ((o = getopt(argc, argv, "hqcsu:e:d")) >= 0) {
		switch (o) {
		case 'h':
			g_print("USAGE:\n");
			g_print("-------------------------------------------------\n");
			g_print("tray_mixer_plus [-h] [-q] [-c] [-s] [-u vol] [-e PCMelement] [-d]\n");
			g_print("where is\n");
			g_print("[-h]            - show this help :)\n");
			g_print("[-q]            - disable Quit menu\n");
			g_print("[-c]            - center window position\n");
			g_print("[-s]            - sleep 5 sec. before go to tray (for icewm)\n");
			g_print("[-u vol]        - force unmute with volume 'vol'\n");
			g_print("[-e PCMelement] - use PCMelement instead of 'Master'\n");
			g_print("[-d]            - dynamic find if soundcard change and not exit  \n");
			g_print("                  if not find any soundcard (wait soundcard) \n");
			return 1;
			break;

		case 'q':
			noquit=1;
			break;

		case 'c':
			wm=1;
			break;

		case 's':
			sleep=1;
			break;

		case 'u':
			unmute=1;
			unvol = optarg;
			g_print("force unmute with %s volume\n", unvol);
			break;

		case 'e':
			name = optarg;
			userelement=1;
			g_print("Use PCM_element=%s instead of 'Master'\n", name);
			break;

		case 'd':
			dynamic=1;
			break;
		}
	}

	g_print("run with -h option for see help\n");

	gtk_init(&argc, &argv);

	if (sleep==1){
	    g_print("sleep ");
	    usleep(5000000);
	    g_print("- end\n");
	}

	mi_ret=mixer_init();
	if (mi_ret==1 && dynamic==0) exit(1);

	icon = gtk_status_icon_new_from_file(ICON_PATH "speaker.png");
	g_signal_connect(G_OBJECT(icon), "activate", G_CALLBACK(click), NULL);

	item = gtk_image_menu_item_new_with_label(_("Quit"));
	item_i = gtk_image_new_from_file(ICON_PATH "quit.png");
	gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(item), item_i);

	item2 = gtk_image_menu_item_new_with_label(_("Extra settings"));
	item2_i = gtk_image_new_from_file(ICON_PATH "setting.png");
	gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(item2), item2_i);

	sep = gtk_separator_menu_item_new();

	menu = gtk_menu_new();
	gtk_widget_show(item2);
	if (noquit==0){
		gtk_widget_show(sep);
		gtk_widget_show(item);
	}

	gtk_menu_shell_append(GTK_MENU_SHELL(menu), item2);

	//add extra menu item
	struct stat st;
	if ( stat(TEST_SOUNDCARD, &st)==0 ) {
	    item3 = gtk_image_menu_item_new_with_label(_("Test soundcard"));
	    item3_i = gtk_image_new_from_file(ICON_PATH "test.png");
	    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(item3), item3_i);
	    gtk_menu_shell_append(GTK_MENU_SHELL(menu), item3);
	    g_signal_connect(G_OBJECT(item3), "activate", G_CALLBACK(testsound), NULL);
	    gtk_widget_show(item3);
	}
	if ( stat(CHANGE_SOUNDCARD, &st)==0 ) {
	    item4 = gtk_image_menu_item_new_with_label(_("Default soundcard"));
	    item4_i = gtk_image_new_from_file(ICON_PATH "default.png");
	    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(item4), item4_i);
	    gtk_menu_shell_append(GTK_MENU_SHELL(menu), item4);
	    g_signal_connect(G_OBJECT(item4), "activate", G_CALLBACK(defsound), NULL);
	    gtk_widget_show(item4);
	}

	gtk_menu_shell_append(GTK_MENU_SHELL(menu), sep);
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);


	g_signal_connect(G_OBJECT(icon), "popup-menu", G_CALLBACK(popup), NULL);
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(quit), NULL);
	g_signal_connect(G_OBJECT(item2), "activate", G_CALLBACK(extra), NULL);

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_decorated(GTK_WINDOW(window), FALSE);
	gtk_container_set_border_width(GTK_CONTAINER(window), 5);
	gtk_window_set_skip_taskbar_hint(GTK_WINDOW(window), TRUE);
	gtk_window_set_skip_pager_hint(GTK_WINDOW(window), TRUE);
	gtk_window_set_type_hint(GTK_WINDOW(window),
					GDK_WINDOW_TYPE_HINT_DIALOG);
	gtk_window_set_default_size(GTK_WINDOW(window), 60, 160);

	hbox = gtk_hbox_new(TRUE, 5);
	gtk_container_add(GTK_CONTAINER(window), hbox);

	/* TODO: make generic function for ch[n] */
	ch[0].vbox = gtk_vbox_new(FALSE, 5);

	ch[0].vscale = gtk_vscale_new(GTK_ADJUSTMENT(
	gtk_adjustment_new(mixer_get(&ch[0]), 0, 100, 0, 0, 0)));
	snprintf(name_gtk, 10, "%s", name);
	ch[0].label= gtk_label_new(name_gtk);
	gtk_scale_set_draw_value(GTK_SCALE(ch[0].vscale), FALSE);
	gtk_range_set_inverted(GTK_RANGE(ch[0].vscale), TRUE);
	gtk_container_add(GTK_CONTAINER(hbox), ch[0].vbox);
	gtk_box_pack_start(GTK_BOX(ch[0].vbox), ch[0].label, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(ch[0].vbox), ch[0].vscale, TRUE, TRUE, 0);

	g_signal_connect((gpointer)ch[0].vscale,
			"value_changed", G_CALLBACK(vol_change), &ch[0]);
	g_signal_connect(ch[0].vscale,
			"scroll-event", G_CALLBACK(scale_scroll), &ch[0]);


	if (snd_mixer_selem_has_playback_switch(elem)) {
		ch[0].mute = gtk_check_button_new_with_label(_("Mute"));

		g_signal_connect((gpointer)ch[0].mute,
				"toggled", G_CALLBACK(mute), &ch[0]);

		gtk_box_pack_end(GTK_BOX(ch[0].vbox), ch[0].mute, FALSE, FALSE, 0);
	}

	g_signal_connect(window,
			"enter-notify-event", G_CALLBACK(cursor_enter), &ch[0]);
	g_signal_connect(ch[0].vbox,
			"enter-notify-event", G_CALLBACK(cursor_enter), &ch[0]);
	g_signal_connect(ch[0].vscale,
			"enter-notify-event", G_CALLBACK(cursor_enter), &ch[0]);

	g_signal_connect(ch[0].label,
			"enter-notify-event", G_CALLBACK(cursor_enter), &ch[0]);
	g_signal_connect(window,
			"leave-notify-event", G_CALLBACK(cursor_leave), &ch[0]);

	//try force unmute
	if (unmute==1){
	    //unmute
	    for (i = 0; i <= SND_MIXER_SCHN_LAST; i++)
		snd_mixer_selem_set_playback_switch(elem, i, 1);

	    unvoldig=atoi(unvol);
	    if (unvoldig>0 && unvoldig<=100){
		mixer_set(&ch[0], unvoldig);
	    }
	}

	update_gui(&ch[0]);

	if (mi_ret==0 ) {
	    gtk_status_icon_set_visible(GTK_STATUS_ICON(icon), TRUE);
	    find_sc=1;
	}else{
	    find_sc=0;
	    //run timer
	    g_timeout_add(1000, mytimer2, NULL);
	}

	gtk_main();

	return 0;
}
